@@ -34,7 +34,7 @@ impl std::str::FromStr for Blob {
     }
 }
 
-#[derive(Debug, Copy, Clone, PartialEq, PartialOrd)]
+#[derive(Copy, Clone, PartialEq, PartialOrd)]
 pub enum QParams {
     MinMax { min: f32, max: f32 },
     ZpScale { zero_point: i32, scale: f32 },
@@ -75,6 +75,23 @@ impl QParams {
             QParams::ZpScale { zero_point, scale } => (*zero_point, *scale),
         }
     }
+
+    pub fn q(&self, f: f32) -> i32 {
+        let (zp, scale) = self.zp_scale();
+        (f / scale) as i32 + zp
+    }
+
+    pub fn dq(&self, i: i32) -> f32 {
+        let (zp, scale) = self.zp_scale();
+        (i - zp) as f32 * scale
+    }
+}
+
+impl std::fmt::Debug for QParams {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        let (zp, scale) = self.zp_scale();
+        write!(f, "Z:{} S:{}", zp, scale)
+    }
 }
 
 #[derive(Debug, Copy, Clone, PartialEq, Eq, Hash, Ord, PartialOrd)]
@@ -96,6 +113,7 @@ pub enum DatumType {
     String,
     QI8(QParams),
     QU8(QParams),
+    QI32(QParams),
     ComplexI16,
     ComplexI32,
     ComplexI64,
@@ -111,9 +129,17 @@ impl DatumType {
         {
             tvec!(*self)
         } else if self.is_complex_float() {
-            [ComplexF16, ComplexF32, ComplexF64].iter().filter(|s| s.size_of() >= self.size_of()).copied().collect()
+            [ComplexF16, ComplexF32, ComplexF64]
+                .iter()
+                .filter(|s| s.size_of() >= self.size_of())
+                .copied()
+                .collect()
         } else if self.is_complex_signed() {
-            [ComplexI16, ComplexI32, ComplexI64].iter().filter(|s| s.size_of() >= self.size_of()).copied().collect()
+            [ComplexI16, ComplexI32, ComplexI64]
+                .iter()
+                .filter(|s| s.size_of() >= self.size_of())
+                .copied()
+                .collect()
         } else if self.is_float() {
             [F16, F32, F64].iter().filter(|s| s.size_of() >= self.size_of()).copied().collect()
         } else if self.is_signed() {
@@ -156,11 +182,17 @@ impl DatumType {
     }
 
     pub fn is_unsigned(&self) -> bool {
-        matches!(self.unquantized(), DatumType::U8 | DatumType::U16 | DatumType::U32 | DatumType::U64)
+        matches!(
+            self.unquantized(),
+            DatumType::U8 | DatumType::U16 | DatumType::U32 | DatumType::U64
+        )
     }
 
     pub fn is_signed(&self) -> bool {
-        matches!(self.unquantized(), DatumType::I8 | DatumType::I16 | DatumType::I32 | DatumType::I64)
+        matches!(
+            self.unquantized(),
+            DatumType::I8 | DatumType::I16 | DatumType::I32 | DatumType::I64
+        )
     }
 
     pub fn is_float(&self) -> bool {
@@ -180,20 +212,22 @@ impl DatumType {
     }
 
     pub fn is_copy(&self) -> bool {
-        *self == DatumType::Bool || self.is_unsigned() || self.is_signed() || self.is_float()
+        *self == DatumType::Bool
+            || self.is_unsigned()
+            || self.is_signed()
+            || self.is_float()
             || self.is_complex()
     }
 
     pub fn is_quantized(&self) -> bool {
-        match self {
-            DatumType::QI8(_) | DatumType::QU8(_) => true,
-            _ => false,
-        }
+        self.qparams().is_some()
     }
 
     pub fn qparams(&self) -> Option<QParams> {
         match self {
-            DatumType::QI8(qparams) | DatumType::QU8(qparams) => Some(*qparams),
+            DatumType::QI8(qparams) | DatumType::QU8(qparams) | DatumType::QI32(qparams) => {
+                Some(*qparams)
+            }
             _ => None,
         }
     }
@@ -202,6 +236,7 @@ impl DatumType {
         match self {
             DatumType::QI8(_) => DatumType::QI8(qparams),
             DatumType::QU8(_) => DatumType::QI8(qparams),
+            DatumType::QI32(_) => DatumType::QI32(qparams),
             _ => *self,
         }
     }
@@ -214,6 +249,7 @@ impl DatumType {
         match self {
             DatumType::QI8(_) => DatumType::I8,
             DatumType::QU8(_) => DatumType::U8,
+            DatumType::QI32(_) => DatumType::I32,
             _ => *self,
         }
     }
@@ -259,6 +295,7 @@ impl DatumType {
             | DatumType::U32
             | DatumType::U64 => Tensor::zero_dt(*self, &[1]).unwrap(),
             DatumType::I8 | DatumType::QI8(_) => tensor0(i8::MIN),
+            DatumType::QI32(_) => tensor0(i32::MIN),
             DatumType::I16 => tensor0(i16::MIN),
             DatumType::I32 => tensor0(i32::MIN),
             DatumType::I64 => tensor0(i64::MIN),
@@ -278,6 +315,7 @@ impl DatumType {
             DatumType::I16 => tensor0(i16::MAX),
             DatumType::I32 => tensor0(i32::MAX),
             DatumType::I64 => tensor0(i64::MAX),
+            DatumType::QI32(_) => tensor0(i32::MAX),
             DatumType::F16 => tensor0(f16(half::f16::MAX)),
             DatumType::F32 => tensor0(f32::MAX),
             DatumType::F64 => tensor0(f64::MAX),
@@ -307,11 +345,11 @@ impl std::str::FromStr for DatumType {
             "String" | "string" => Ok(DatumType::String),
             "TDim" | "tdim" => Ok(DatumType::TDim),
             "ComplexI16" | "complexi16" => Ok(DatumType::ComplexI16),
-            "ComplexI32" | "complexi32"  => Ok(DatumType::ComplexI32),
-            "ComplexI64" | "complexi64"  => Ok(DatumType::ComplexI64),
-            "ComplexF16" | "complexf16"  => Ok(DatumType::ComplexF16),
-            "ComplexF32" | "complexf32"  => Ok(DatumType::ComplexF32),
-            "ComplexF64" | "complexf64"  => Ok(DatumType::ComplexF64),
+            "ComplexI32" | "complexi32" => Ok(DatumType::ComplexI32),
+            "ComplexI64" | "complexi64" => Ok(DatumType::ComplexI64),
+            "ComplexF16" | "complexf16" => Ok(DatumType::ComplexF16),
+            "ComplexF32" | "complexf32" => Ok(DatumType::ComplexF32),
+            "ComplexF64" | "complexf64" => Ok(DatumType::ComplexF64),
             _ => anyhow::bail!("Unknown type {}", s),
         }
     }
@@ -351,10 +389,15 @@ pub trait ClampCast: PartialOrd + Copy + 'static {
     #[inline(always)]
     fn clamp_cast<O>(self) -> O
     where
-        Self: AsPrimitive<O>,
-        O: AsPrimitive<Self> + num_traits::Bounded,
+        Self: AsPrimitive<O> + Datum,
+        O: AsPrimitive<Self> + num_traits::Bounded + Datum,
     {
-        num_traits::clamp(self, O::min_value().as_(), O::max_value().as_()).as_()
+        // this fails if we're upcasting, in which case clamping is useless
+        if O::min_value().as_() < O::max_value().as_() {
+            num_traits::clamp(self, O::min_value().as_(), O::max_value().as_()).as_()
+        } else {
+            self.as_()
+        }
     }
 }
 impl<T: PartialOrd + Copy + 'static> ClampCast for T {}
