@@ -57,6 +57,7 @@ impl Hash for Tensor {
                 Blob => self.as_slice_unchecked::<crate::datum::Blob>().hash(state),
                 QI8(_) => self.as_slice_unchecked::<i8>().hash(state),
                 QU8(_) => self.as_slice_unchecked::<u8>().hash(state),
+                QI32(_) => self.as_slice_unchecked::<i32>().hash(state),
                 ComplexI16 => self.as_slice_unchecked::<Complex<i16>>().hash(state),
                 ComplexI32 => self.as_slice_unchecked::<Complex<i32>>().hash(state),
                 ComplexI64 => self.as_slice_unchecked::<Complex<i64>>().hash(state),
@@ -608,7 +609,15 @@ impl Tensor {
     /// `force_full` will force the tensor to be dump in full even if it is big.
     pub fn dump(&self, force_full: bool) -> anyhow::Result<String> {
         unsafe fn dump_t<D: Datum>(tensor: &Tensor, n: usize) -> String {
-            tensor.as_slice_unchecked::<D>()[0..n].iter().join(", ")
+            if let Some(qp) = tensor.datum_type().qparams() {
+                let integers = tensor.cast_to::<i32>().unwrap();
+                integers.as_slice_unchecked::<i32>()[0..n]
+                    .iter()
+                    .map(|x| format!("[{}]({})", x, qp.dq(*x)))
+                    .join(", ")
+            } else {
+                tensor.as_slice_unchecked::<D>()[0..n].iter().join(", ")
+            }
         }
         unsafe {
             let trunc = self.len() > 12 && !force_full;
@@ -955,8 +964,8 @@ impl Tensor {
                                 .iter()
                                 .zip(result.as_slice_mut_unchecked::<$typ>().iter_mut())
                                 .for_each(|(&s, d)| {
-                                    *d = (d_zp as i16
-                                        + scale_by(s as i16 - s_zp as i16, s_scale / d_scale))
+                                    *d = (d_zp as i32
+                                        + scale_by(s as i32 - s_zp as i32, s_scale / d_scale))
                                     .clamp_cast()
                                 });
                             return Ok(Cow::Owned(result));
@@ -1021,12 +1030,18 @@ impl Tensor {
 
                 q_f!(f32, i8, round_ties_to_even);
                 q_f!(f32, u8, round_ties_to_even);
+                q_f!(f32, i32, round_ties_to_even);
                 q_f!(i8, f32, |f| f);
                 q_f!(u8, f32, |f| f);
+                q_f!(i32, f32, |f| f);
 
                 if dst_dt.is_quantized() && self.datum_type().is_quantized() {
                     q_f!(u8, i8, round_ties_to_even);
                     q_f!(i8, u8, round_ties_to_even);
+                    q_f!(i32, u8, round_ties_to_even);
+                    q_f!(i32, i8, round_ties_to_even);
+                    q_f!(u8, i32, round_ties_to_even);
+                    q_f!(i8, i32, round_ties_to_even);
                 }
 
                 q_n!(i8, i32);
